/*
 * PDCA.hpp
 *
 *  Created on: 2012-4-8
 *      Author: Andygongyb
 */

#ifndef PDCA_HPP_
#define PDCA_HPP_

#include <avr32/io.h>
#include <cstddef>
#include <cstring>
#include <stdint.h>

extern "C" {
#include "intc.h"
#include "pdca.h"
}

#include "Exception.hpp"
#include "Initialization.hpp"

class CPDCA_Peripheral;

typedef class CPDCA_exception: public CException {
	static const char *m_already_attached;
	static const char *m_all_in_use;
public:

	typedef enum {
		ALREADY_ATTACHED = -1, ALL_IN_USE = -2,
	} errno_t;

	CPDCA_exception(errno_t errno) :
		CException(NULL, 0) {
		const char *name = NULL;
		switch (errno) {
		case ALREADY_ATTACHED:
			name = m_already_attached;
			break;
		case ALL_IN_USE:
			name = m_all_in_use;
			break;
		}
		strcpy(m_name, name);
		m_errno = errno;
	}
}*pCPDCA_exception;

class CPDCAManager;

typedef class CPDCA {
public:

	typedef enum {
		BYTE = 0, HALF_WORD = 1, WORD = 2,
	} mode_t;

	typedef enum {
		ADCIFA0_RX = 0,
		ADCIFA1_RX = 1,
		USART0_RX = 2,
		USART1_RX = 3,
		USART2_RX = 4,
		USART3_RX = 5,
		TWIM0_RX = 6,
		TWIM1_RX = 7,
		TWIS0_RX = 8,
		TWIS1_RX = 9,
		SPI0_RX = 10,
		SPI1_RX = 11,
		AW_RX = 12,
		USART0_TX = 13,
		USART1_TX = 14,
		USART2_TX = 15,
		USART3_TX = 16,
		TWIM0_TX = 17,
		TWIM1_TX = 18,
		TWIS0_TX = 19,
		TWIS1_TX = 20,
		SPI0_TX = 21,
		SPI1_TX = 22,
		DACIFB00_TX = 23,
		DACIFB01_TX = 24,
		DACIFB10_TX = 25,
		DACIFB11_TX = 26,
		PWM_TX = 27,
		AW_TX = 28,
		USART4_RX = 31,
		TWIM2_RX = 32,
		TWIS2_RX = 33,
		USART4_TX = 34,
		TWIM2_TX = 35,
		TWIS2_TX = 36,
	} module_t;

	typedef enum {
		PDCA0 = 0,
		PDCA1,
		PDCA2,
		PDCA3,
		PDCA4,
		PDCA5,
		PDCA6,
		PDCA7,
		PDCA8,
		PDCA9,
		PDCA10,
		PDCA11,
		PDCA12,
		PDCA13,
		PDCA14,
		PDCA15,
	} channel_t;

	typedef struct {
		mode_t size;
		module_t module;
	} property_t;

protected:
	volatile avr32_pdca_channel_t * const m_channel;
	CPDCA_Peripheral *m_peripheral;
	property_t *m_pproperty;
	const uint8_t m_irq;
	static const uint8_t irq_map[16];

public:

	CPDCA(channel_t channel) :
		m_channel(&AVR32_PDCA.channel[channel]), m_peripheral(NULL), m_irq(
				irq_map[channel]) {
		CInit::instance();
		m_channel->CR.tdis = 1; // disable the channel by default
	}

	virtual ~CPDCA(void) {

	}

	void set_property(property_t &property) {
		m_channel->MR.size = static_cast<uint32_t> (property.size);
		m_channel->PSR.pid = static_cast<uint32_t> (property.module);
		memcpy(static_cast<void *> (m_pproperty),
				static_cast<const void *> (&property), sizeof(property_t));
	}

	void get_property(property_t &property) {
		memcpy(static_cast<void *> (&property),
				static_cast<const void *> (m_pproperty), sizeof(property_t));
	}

	volatile avr32_pdca_channel_t * const get_channel(void) {
		return m_channel;
	}

	void set_handler(void(*handler)(void)) {
		// Register DMA interrupt handler on level 3
		bool int_sta = Is_global_interrupt_enabled();
		Disable_global_interrupt();
		INTC_register_interrupt(handler, m_irq, AVR32_INTC_INT3);
		if (int_sta)
			Enable_global_interrupt();
	}

	void set_next_mem_loc(void* loc, uint16_t length) {
		/*
		 * we can always use MARR instead of MAR because
		 * the value in MARR will be copied to MAR immediately
		 * after MAR is empty
		 */
		m_channel->marr = (unsigned long) loc;
		m_channel->tcrr = length; // same for TCRR and TCR here
	}

	void on(void) {
		m_channel->CR.ten = 1;
	}

	void off(void) {
		/*
		 * we may want to clear MAR and TCR here
		 * or not?
		 */
		m_channel->CR.tdis = 1;
	}

	void clear_err(void) {
		m_channel->CR.eclr = 1;
	}

	void en_err_int(void) {
		m_channel->IER.terr = 1;
	}

	void dis_err_int(void) {
		m_channel->IDR.terr = 1;
	}

	void dis_all_int(void) {
		m_channel->idr = ~0UL;
	}

	void en_trans_fin_int(void) {
		m_channel->IER.trc = 1;
	}

	void dis_trans_fin_int(void) {
		m_channel->IDR.trc = 1;
	}

	void en_reload_count_zero_int(void) {
		m_channel->IER.rcz = 1;
	}

	void dis_reload_count_zero_int(void) {
		m_channel->IDR.rcz = 1;
	}

	friend class CPDCA_Peripheral;
	friend class CPDCAManager;

}*pCPDCA;

template<CPDCA::channel_t N>
class CPDCA_ch: public CPDCA {
private:
	static CPDCA_ch *m_instance;
	static property_t m_property;

	CPDCA_ch(void) :
		CPDCA(N) {
		m_pproperty = &m_property;
	}

public:
	static CPDCA_ch *instance(void) {
		if (!m_instance)
			m_instance = new CPDCA_ch();
		return m_instance;
	}

	virtual ~CPDCA_ch(void) {
		off();
		m_instance = NULL;
	}
};

typedef class CPDCAManager {

	static CPDCAManager *m_instance;
	CPDCA *m_pdca[16];

	CPDCAManager(void) {
		m_pdca[0] = CPDCA_ch<CPDCA::PDCA0>::instance();
		m_pdca[1] = CPDCA_ch<CPDCA::PDCA1>::instance();
		m_pdca[2] = CPDCA_ch<CPDCA::PDCA2>::instance();
		m_pdca[3] = CPDCA_ch<CPDCA::PDCA3>::instance();
		m_pdca[4] = CPDCA_ch<CPDCA::PDCA4>::instance();
		m_pdca[5] = CPDCA_ch<CPDCA::PDCA5>::instance();
		m_pdca[6] = CPDCA_ch<CPDCA::PDCA6>::instance();
		m_pdca[7] = CPDCA_ch<CPDCA::PDCA7>::instance();
		m_pdca[8] = CPDCA_ch<CPDCA::PDCA8>::instance();
		m_pdca[9] = CPDCA_ch<CPDCA::PDCA9>::instance();
		m_pdca[10] = CPDCA_ch<CPDCA::PDCA10>::instance();
		m_pdca[11] = CPDCA_ch<CPDCA::PDCA11>::instance();
		m_pdca[12] = CPDCA_ch<CPDCA::PDCA12>::instance();
		m_pdca[13] = CPDCA_ch<CPDCA::PDCA13>::instance();
		m_pdca[14] = CPDCA_ch<CPDCA::PDCA14>::instance();
		m_pdca[15] = CPDCA_ch<CPDCA::PDCA15>::instance();
	}

public:

	static CPDCAManager *instance(void) {
		if (!m_instance)
			m_instance = new CPDCAManager();
		return m_instance;
	}

	~CPDCAManager(void) {
		m_instance = NULL;
	}

	CPDCA *get(void) {
		CPDCA **the_pdca = m_pdca;
		uint8_t count = 0;
		while ((*the_pdca)->m_peripheral) {
			count++;
			if (count > 15) {
				throw CPDCA_exception(CPDCA_exception::ALL_IN_USE);
			}
			the_pdca++;
		}
		return *the_pdca;
	}

}*pCPDCAManager;

extern CPDCAManager *the_pdca_manager;

typedef class CPDCA_Peripheral {
protected:

	CPDCA *m_pdca;

	void attach_pdca(CPDCA *&the_pdca) {
		if (!the_pdca) {
			the_pdca = CPDCAManager::instance()->get();
			the_pdca->m_peripheral = this;
		}
	}

	void detach_pdca(CPDCA *&the_pdca) {
		if (the_pdca) {
			the_pdca->dis_all_int();
			the_pdca->off();
			the_pdca->m_peripheral = NULL;
			the_pdca = NULL;
		}
	}

	virtual ~CPDCA_Peripheral(void) {
		detach_pdca(m_pdca);
	}

}*pCPDCA_Peripheral;

#endif /* PDCA_HPP_ */
