/*
 * USB_CDC.cpp
 *
 *  Created on: Nov 27, 2011
 *      Author: Andygongyb
 */

#include <string.h>
#include "USB_CDC.hpp"

using namespace usb_cdc_names;

uint8_t usb_cdc_names::line_coding_dummy[7] = { 0x80, 0x25, 0x00, 0x00, 0x00,
		0x00, 0x08, };

CUSB_CDC *CUSB_CDC::m_instance = NULL;

CUSB_CDC::CUSB_CDC(void) {
	pwr(false);
}

void CUSB_CDC::pwr(bool if_on) {
	if (if_on) {

		detach(); // make sure the device is disconnected

		m_state = DETACHED;
		m_request_state = IDLE;

		USBCON = 0; // reset the USB controller
		UHWCON = _BV(UVREGE); // enable pad regulator

		CInit::instance()->pll(true); // enable PLL

		USBCON = _BV(USBE);

		reset_endpoints();

		// enable the usb controller, its clock, VBUS detection, and the corresponding interrupt.
		USBCON = _BV(USBE) | _BV(OTGPADE) | _BV(VBUSTE);

		UDIEN = _BV(SUSPE) | _BV(EORSTE);

	} else {

		detach();
		USBCON = _BV(FRZCLK); // disable the controller and freeze the clock

		CInit::instance()->pll(false); // disable PLL

		UHWCON = 0; // disable pad regulator

	}
}

uint8_t CUSB_CDC::write_bank(uint8_t *buff, uint8_t num) {

	while (!m_tx_empty) {
		wdt_reset();
	}

	portENTER_CRITICAL();

	UENUM = 0x01; // select EP1 for write

	for (uint8_t *p = buff; num; num--) {
		UEDATX = *p;
		p++;
		if (UEBCLX == 64) {
			break;
		}
	}

	portEXIT_CRITICAL();

	return (64 - UEBCLX); // return the remain capacity for write
}

uint8_t CUSB_CDC::read_bank(uint8_t *buff) {
	if (!m_rx_count) {
		return 0;
	}
	portENTER_CRITICAL();
	memcpy(buff, m_rx_buffer, m_rx_count);
	uint8_t tmp = m_rx_count;
	m_rx_count = 0;
	//UENUM = 0x02;
	//UEINTX &= ~_BV(FIFOCON);
	portEXIT_CRITICAL();
	return tmp;
}

void CUSB_CDC::send(void) {

	while (!m_tx_empty) {
		wdt_reset();
	}

	portENTER_CRITICAL();

	UENUM = 0x01; // select EP1 for write

	m_tx_empty = false;
	UEINTX &= ~_BV(FIFOCON);

	portEXIT_CRITICAL()
	;

}

int8_t CUSB_CDC::read_request(void) {
	UENUM = 0;

	if (UEBCLX != 8) {
		return -1;
	}

	for (uint8_t i = 0; i < 8; i++) {
		m_request.buffer[i] = UEDATX;
	}

	return 0;

}

void CUSB_CDC::enable_endpoint_0(void) {
	UENUM = 0x00;
	UECONX = _BV(EPEN); // enable EP 0
	UECFG0X = 0x00; // set EP0 as control OUT
	UECFG1X = _BV(EPSIZE1) | _BV(EPSIZE0) | _BV(ALLOC); // set EP0 to 64 byte with single bank
	UEIENX = _BV(RXSTPE) | _BV(RXOUTE) | _BV(TXINE); // enable the SETUP, IN, and OUT interrupts
}

void CUSB_CDC::enable_endpoints(void) {

	UENUM = 0x01; // select EP1
	UECONX = _BV(EPEN); // enable EP1
	UECFG0X = _BV(EPTYPE1) | _BV(EPDIR); // set EP1 as bulk IN
	UECFG1X = _BV(EPSIZE1) | _BV(EPSIZE0) | _BV(EPBK0) | _BV(ALLOC); // set EP1 size as 64 bytes with double bank
	UEIENX = _BV(TXINE); // enable the IN interrupt

	UENUM = 0x02; // select EP2
	UECONX = _BV(EPEN); // enable EP2
	UECFG0X = _BV(EPTYPE1); // set EP2 as bulk OUT
	UECFG1X = _BV(EPSIZE1) | _BV(ALLOC); // set EP2 size as 32 bytes
	UEIENX = _BV(RXOUTE); // enable the OUT interrupt


	UENUM = 0x03; // select EP3
	UECONX = _BV(EPEN); // enable EP3
	UECFG0X = _BV(EPTYPE1) | _BV(EPTYPE0) | _BV(EPDIR); // set EP3 as interrupt IN
	UECFG1X = _BV(EPSIZE1) | _BV(EPSIZE0) | _BV(ALLOC); // set EP3 size as 64 bytes
	UEIENX = _BV(TXINE); // enable the IN interrupt



}

uint8_t CUSB_CDC::respond_request(const uint8_t *buffer, uint8_t len) {
	uint8_t ret = 0;
	if (len > 64) {
		ret = len - 64;
		len = 64;
	}

	UENUM = 0x00; // select EP0
	for (uint8_t i = 0; i < len; i++) {
		UEDATX = buffer[i];
	}

	UEINTX &= ~_BV(TXINI); // send the response
	return ret;
}

void USB_GEN_vect(void) {

	CUSB_CDC *usb = CUSB_CDC::instance();

	// transition on VBUS
	if (USBINT) {

		if (USBSTA & _BV(VBUS)) {
			usb->attach();
			usb->m_state = CUSB_CDC::ATTACHED;
		} else {
			usb->detach();
			usb->m_state = CUSB_CDC::DETACHED;
		}

		USBINT = 0; // clear VBUS transition interrupt flag
		return; // we are not likely having other interrupts simultaneously
	}

	uint8_t status = UDINT;

	static CUSB_CDC::state_t last_state;

	if (status & _BV(SUSPI)) { // suspended
		last_state = usb->m_state;
		usb->m_state = CUSB_CDC::SUSPENDED;
		USBCON |= _BV(FRZCLK); // freeze the clock
		UDIEN |= _BV(WAKEUPE); // enable the wake-up interrupt
	}

	if (status & _BV(WAKEUPI)) { // woke up
		usb->m_state = last_state;
		UDIEN &= ~_BV(WAKEUPE); // disable the wake-up interrupt
		USBCON &= ~_BV(FRZCLK); // resume the clock
	}

	if (status & _BV(EORSTI)) { // end of reset
		usb->enable_endpoint_0();
		usb->enable_endpoints();
		usb->m_state = CUSB_CDC::ATTACHED;
	}

	UDINT = 0; // clear all USB device interrupt flags

}

void USB_COM_vect(void) {

	CUSB_CDC *usb = CUSB_CDC::instance();
	uint8_t status;

	static uint8_t data_remain, len;
	static const uint8_t *ptr;
	
	while (UEINT) {
	
		UENUM = 0x02; // set to EP2
		if (UEINTX & _BV(RXOUTI)) {
			UEINTX &= ~_BV(RXOUTI);
			uint8_t tmp = UEBCLX;
			for (uint8_t i = 0; i < tmp; i++) {
				usb->m_rx_buffer[i + usb->m_rx_count] = UEDATX;
			}
			usb->m_rx_count += tmp;
			UEINTX &= ~_BV(FIFOCON);
		}

		UENUM = 0x03; // set to EP3
		if (UEINTX & _BV(TXINI)) {
			UEINTX &= ~_BV(TXINI);
			UEINTX &= ~_BV(FIFOCON);
		}

		UENUM = 0x01; // set to EP1
		if (UEINTX & _BV(TXINI)) {
			UEINTX &= ~_BV(TXINI);
			usb->m_tx_empty = true;
		}

		UENUM = 0x00; // set to EP 0
		status = UEINTX;
		if (status & _BV(RXSTPI)) { // received a setup signal
			if (usb->read_request() == -1) {
				return; // read the command
			}

			UEINTX &= ~_BV(RXSTPI); // clear the flag

			if (usb->m_request.request_type.dir) { // input request
				usb->m_request_state = CUSB_CDC::IN;
				switch (usb->m_request.request_type.type) {
				case 0: // standard request
					switch (usb->m_request.request) {
					case GET_CONFIGURATION:
						return;
					case GET_DESCRIPTOR:
						switch (usb->m_request.value >> 8) {
						case DEVICE_DESCRIPTOR:
							ptr = device_descriptor;
							data_remain = (device_descriptor[0]
									> usb->m_request.len) ? usb->m_request.len
									: device_descriptor[0];
							len = data_remain;
							data_remain = usb->respond_request(ptr, data_remain);
							return;
						case CONFIGURATION_DESCRIPTOR:
							ptr = conf_descriptor;
							data_remain = conf_descriptor[3] * 256
									+ conf_descriptor[2];
							data_remain
									= (data_remain > usb->m_request.len) ? usb->m_request.len
											: data_remain;
							len = data_remain;
							data_remain = usb->respond_request(ptr, data_remain);
							return;
						case STRING_DESCRIPTOR:
							switch (usb->m_request.value & 0xFF) {

							case 0: // language ID
								ptr = language_id;
								data_remain
										= (language_id[0] > usb->m_request.len) ? usb->m_request.len
												: language_id[0];
								break;

							case 1: // manufacturer string
								ptr = manufacturer_string_descriptor;
								data_remain = (manufacturer_string_descriptor[0]
										> usb->m_request.len) ? usb->m_request.len
										: manufacturer_string_descriptor[0];
								break;

							case 2: // product string
								ptr = product_string_descriptor;
								data_remain = (product_string_descriptor[0]
										> usb->m_request.len) ? usb->m_request.len
										: product_string_descriptor[0];
								break;

							case 3: // serial number
								ptr = serial_number_string_descriptor;
								data_remain = (serial_number_string_descriptor[0]
										> usb->m_request.len) ? usb->m_request.len
										: serial_number_string_descriptor[0];
								break;

							default:
								UEINTX &= ~_BV(TXINI); // send a 0 length package
								return;
							}
							len = data_remain;
							data_remain = usb->respond_request(ptr, data_remain);
							return;
						default:
							return;
						}

					case GET_INTERFACE:
					case GET_STATUS:
					case SYNCH_FRAME:
					default:
						return;
					}

				case 1: // class request

					switch (usb->m_request.request) {
					case GET_LINE_CODING:
						ptr = const_cast<const uint8_t*> (line_coding_dummy);
						data_remain = 7;
						len = data_remain;
						data_remain = usb->respond_request(ptr, data_remain);
						return;

					case SERIAL_STATE:
					default:
						UEINTX &= ~_BV(TXINI); // send a 0 length package
						return;
					}

				default:
					return;

				}
			} else {

				usb->m_request_state = CUSB_CDC::OUT;
				switch (usb->m_request.request_type.type) {
				case 0: // standard request

					switch (usb->m_request.request) {
					case SET_ADDRESS:
						UDADDR = static_cast<uint8_t> (usb->m_request.value & 0xFF); // set the address
						usb->m_state = CUSB_CDC::ASSIGN_ADDRESS;
						UEINTX &= ~_BV(TXINI); // send a 0 length package
						return;

					case CLEAR_FEATURE:
					case SET_DESCRIPTOR:
					case SET_FEATURE:
					case SET_INTERFACE:
						return;

					case SET_CONFIGURATION: // we only have one configuration here
						usb->m_tx_empty = true; // from here we allow user to fill the bank
					default:
						UEINTX &= ~_BV(TXINI); // send a 0 length package
						return;
					}

				case 1: // class request
					switch (usb->m_request.request) {

					case SET_LINE_CODING:
						usb->m_request_state = CUSB_CDC::SET_LINE;
						return;

					case SET_IDLE:
					case SET_CONTROL_LINE_STATE:
					default:
						UEINTX &= ~_BV(TXINI); // send a 0 length package
						return;
					}

				case 2: // manufacturer request
					return;

				default:
					return;

				}
			}

		}

		if (status & _BV(RXOUTI)) { // received an OUT signal

			if (usb->m_request_state == CUSB_CDC::SET_LINE) {

				if (UEBCLX == 7) {
					for (uint8_t i = 0; i < 7; i++) {
						line_coding_dummy[i] = UEDATX;
					}
					UEINTX &= ~_BV(RXOUTI);
					usb->m_request_state = CUSB_CDC::IDLE;
					UEINTX &= ~_BV(TXINI); // send a 0 length package
					return;
				}
			}

			UEINTX &= ~_BV(RXOUTI);
			if (status & _BV(TXINI)) {
				usb->m_request_state = CUSB_CDC::IDLE;
			}
			return;
		}

		if (status & _BV(TXINI)) { // received an IN signal

			if (usb->m_state == CUSB_CDC::ASSIGN_ADDRESS) {
				UDADDR |= _BV(ADDEN);
				usb->m_if_connected = true;
				usb->m_state = CUSB_CDC::ADDRESSED;
				usb->m_request_state = CUSB_CDC::IDLE;
			}

			if (data_remain) {
				ptr += len - data_remain;
				data_remain = usb->respond_request(ptr, data_remain);
			}
			return;
		}
	
	}	

}
